﻿// ===============================================================================
//  解决方案：置换型模板引擎
//  项目名称：ReplacementTemplateEngine
//  文件名称：TemplateLexer.cs
//  代码作者：网鸟老兵团(Me.YMind)
//  创建时间：2012-04-04 12:35
// ===============================================================================
//  Copyright © Htmlbird.Net. All rights reserved .
//  官方网站：http://www.htmlbird.net/
//  技术论坛：http://bbs.htmlbird.net/
//  版权所有：网鸟IT技术论坛 彦铭工作室
// ===============================================================================

using System;
using System.Collections.Generic;

namespace ReplacementTemplateEngine.Parser
{
	/// <summary>
	/// 表示模板词法分析器。
	/// </summary>
	internal sealed class TemplateLexer : ITemplateLexer
	{
		private const char _EOF = (char)0;

		private readonly string _templateString;
		private int _column;
		private LexerMode _currentMode;
		private int _line;
		private Stack<LexerMode> _modes;
		private int _position;

		private int _saveColumn;
		private int _saveLine;
		private int _savePos;

		/// <summary>
		/// 初始化 <see cref="TemplateLexer"/> 类的新实例。
		/// </summary>
		/// <param name="templateString">包含模板代码的字符串。</param>
		public TemplateLexer(string templateString)
		{
			if (templateString == null) throw new ArgumentNullException("templateString");
			if (templateString.Length == 0) throw new ArgumentOutOfRangeException("templateString");

			this._templateString = templateString;
			this._Reset();
		}

		#region ITemplateLexer Members

		/// <summary>
		/// 继续分析下一条词汇，并返回分析结果。
		/// </summary>
		/// <returns>Token</returns>
		public Token Next()
		{
			switch (this._currentMode)
			{
				case LexerMode.Text:
					return this._NextText();

				case LexerMode.Label:
					return this._NextLabel();

				case LexerMode.FormatString:
					return this._NextFormatString();

				default:
					throw new Exception(String.Format(@"在执行模板词法解析的过程中遇到未知的词法模式 ""{0}""。行：{1}，列：{2}。", this._currentMode, this._line, this._column));
			}
		}

		#endregion

		private Token _NextText()
		{
			this._StartRead();

			startRead:
			switch (this._La(0))
			{
				case _EOF:
					if (this._savePos == this._position) return this._CreateToken(TokenKind.EOF);
					break;

				case '{': // _NextLabel()
					if (this._La(1) == '{')
					{
						this._Consume();

						goto startRead;
					}

					if (this._savePos == this._position)
					{
						this._Consume(1);
						this._EnterMode(LexerMode.Label);

						return this._CreateToken(TokenKind.LeftBracket);
					}

					break;

				case '\n':
				case '\r':
					this._ReadWhitespace();

					goto startRead;

				default:
					this._Consume();

					goto startRead;
			}

			return this._CreateToken(TokenKind.Text);
		}

		private Token _NextLabel()
		{
			this._StartRead();

			startRead:
			switch (this._La(0))
			{
				case _EOF:
					if (this._savePos == this._position) return this._CreateToken(TokenKind.EOF);
					break;

				case '}': // 跳出模式
					if (this._savePos == this._position)
					{
						this._Consume(1);
						this._LeaveMode();

						return this._CreateToken(TokenKind.RightBracket);
					}

					break;

				case ':': // _NextFormatString()
					if (this._savePos == this._position)
					{
						this._Consume(1);
						// 直接退出Label模式
						this._LeaveMode();
						this._EnterMode(LexerMode.FormatString);

						return this._CreateToken(TokenKind.FormatStringPreamble);
					}

					break;

				case ' ':
				case '\t':
				case '\n':
				case '\r':
					this._ReadWhitespace();

					goto startRead;

				default:
					this._Consume();

					goto startRead;
			}

			return this._CreateToken(TokenKind.Label);
		}

		private Token _NextFormatString()
		{
			this._StartRead();

			startRead:
			switch (this._La(0))
			{
				case _EOF:
					if (this._savePos == this._position) return this._CreateToken(TokenKind.EOF);
					break;

				case '}': // 跳出模式
					if (this._savePos == this._position)
					{
						this._Consume(1);
						this._LeaveMode();

						return this._CreateToken(TokenKind.RightBracket);
					}

					break;

				case ' ':
				case '\t':
				case '\n':
				case '\r':
					this._ReadWhitespace();

					goto startRead;

				default:
					this._Consume();

					goto startRead;
			}

			return this._CreateToken(TokenKind.FormatString);
		}

		private void _EnterMode(LexerMode mode)
		{
			this._modes.Push(this._currentMode);
			this._currentMode = mode;
		}

		private void _LeaveMode() { this._currentMode = this._modes.Pop(); }

		private void _Reset()
		{
			this._modes = new Stack<LexerMode>();
			this._currentMode = LexerMode.Text;
			this._modes.Push(this._currentMode);

			this._line = 1;
			this._column = 1;
			this._position = 0;
		}

		private void _NewLine()
		{
			this._line++;
			this._column = 1;
		}

		private void _StartRead()
		{
			this._saveLine = this._line;
			this._saveColumn = this._column;
			this._savePos = this._position;
		}

		private char _La(int count)
		{
			var i = this._position + count;

			return i >= this._templateString.Length ? _EOF : this._templateString[i];
		}

		private void _Consume()
		{
			this._position++;
			this._column++;
		}

		private void _Consume(int count)
		{
			if (count <= 0) throw new ArgumentOutOfRangeException("count", "count 必须大于等于 0。");

			while (count > 0)
			{
				this._Consume();

				count--;
			}
		}

		private Token _CreateToken(TokenKind kind)
		{
			var tokenData = this._templateString.Substring(this._savePos, this._position - this._savePos);

			if (kind == TokenKind.Text) tokenData = tokenData.Replace("\"\"", "\"");

			return new Token(kind, tokenData, this._saveLine, this._saveColumn);
		}

		private void _ReadWhitespace()
		{
			while (true)
			{
				var ch = this._La(0);

				switch (ch)
				{
					case ' ':
					case '\t':
						this._Consume();
						break;

					case '\n':
						this._Consume();
						this._NewLine();

						break;

					case '\r':
						this._Consume();

						if (this._La(0) == '\n') this._Consume();

						this._NewLine();

						break;

					default:
						return;
				}
			}
		}
	}
}
