﻿// ===============================================================================
//  解决方案：置换型模板引擎
//  项目名称：ReplacementTemplateEngine
//  文件名称：TemplateEngine.cs
//  代码作者：网鸟老兵团(Me.YMind)
//  创建时间：2012-04-03 20:58
// ===============================================================================
//  Copyright © Htmlbird.Net. All rights reserved .
//  官方网站：http://www.htmlbird.net/
//  技术论坛：http://bbs.htmlbird.net/
//  版权所有：网鸟IT技术论坛 彦铭工作室
// ===============================================================================

using System;
using System.Collections.Generic;
using System.Text;

namespace ReplacementTemplateEngine.bak2
{
	/// <summary>
	/// 一个简单的置换型模版引擎。
	/// </summary>
	public sealed class TemplateEngine : ITemplateEngine
	{
		private const char _LEFT_BRACKET = '{';
		private const char _RIGHT_BRACKET = '}';
		private const char _FORMAT_STRING_PREAMBLE = ':';

		private readonly Stack<LexerMode> _lexerModes = new Stack<LexerMode>();
		private readonly List<Token> _tokens = new List<Token>();
		private readonly Dictionary<string, object> _variables = new Dictionary<string, object>();
		private int _column;
		private LexerMode _currentMode;
		private int _line;
		private int _postion;

		private int _savedColumn;
		private int _savedLine;
		private int _savedPostion;

		private string _templateString;

		/// <summary>
		/// 获取 Token 列表。
		/// </summary>
		public List<Token> Tokens
		{
			get
			{
				// 此处仅作为显示测试之用，为了防止从外部更改 _tokens 的值
				// 这里在每次调用时生成新的集合副本，下一课我们将会对代码进行重构
				return new List<Token>(this._tokens);
			}
		}

		#region ITemplateEngine Members

		/// <summary>
		/// 解析模板。
		/// </summary>
		/// <param name="templateString">包含模板内容的字符串。</param>
		public void Parser(string templateString)
		{
			if (String.IsNullOrEmpty(templateString)) return;

			this._line = 1;
			this._column = 1;
			this._postion = 0;
			this._savedPostion = 0;
			this._currentMode = LexerMode.Text;
			this._tokens.Clear();
			this._lexerModes.Clear();

			this._templateString = templateString;

			var templateStringLength = this._templateString.Length;

			for (var index = 0; index < templateStringLength; index++)
			{
				var c = templateString[index];

				this._postion++;
				this._column++;

				switch (c)
				{
					case _LEFT_BRACKET:
						if (this._currentMode == LexerMode.Label) continue;

						// 进入Label模式
						this._postion--;
						this._EnterMode(LexerMode.Label);
						this._postion++;
						this._tokens.Add(this._CreateToken(TokenKind.LeftBracket, "{"));
						this._Save();

						break;

					case _RIGHT_BRACKET:
						this._postion--;

						if (this._currentMode == LexerMode.FormatString) this._LeaveMode();

						this._LeaveMode();
						this._postion++;
						this._tokens.Add(this._CreateToken(TokenKind.RightBracket, "}"));
						this._Save();

						break;

					case _FORMAT_STRING_PREAMBLE:
						// 当前模式为Label的时候，遇到':'才进入FormatString模式。
						if (this._currentMode == LexerMode.FormatString) continue;

						if (this._currentMode == LexerMode.Label)
						{
							this._postion--;
							this._EnterMode(LexerMode.FormatString);
							this._postion++;

							this._tokens.Add(this._CreateToken(TokenKind.FormatStringPreamble, ":"));
							this._Save();
						}
						break;

					case '\n': // 换行咯！
						this._line++;
						break;
				}
			}

			if (this._postion <= 0) return;

			var text = this._templateString.Substring(this._savedPostion, this._postion - this._savedPostion);
			var token = this._CreateToken(text);

			this._tokens.Add(token);

			this._Save();
		}

		/// <summary>
		/// 设定变量标记的值。
		/// </summary>
		/// <param name="key">键名。</param>
		/// <param name="value">值。</param>
		public void SetValue(string key, object value)
		{
			// 就这么简单
			this._variables[key] = value;
		}

		/// <summary>
		/// 处理模板并输出结果。
		/// </summary>
		/// <returns>返回包含业务数据的字符串。</returns>
		public string Process()
		{
			var result = new StringBuilder();

			for (var index = 0; index < this._tokens.Count; index++)
			{
				var token = this._tokens[index];

				switch (token.Kind)
				{
					case TokenKind.Label:
						string value;

						// 具体的Token流是：
						// Label					= CreationTime
						// FormatStringPreamble		= :
						// FormatString				= yyyy年MM月dd日 HH:mm:ss
						// 因此这里减去2个索引值检查操作范围
						if (index < this._tokens.Count - 2)
						{
							// 实现自定义格式化字符串
							var nextToken = this._tokens[index + 2];

							if (nextToken.Kind == TokenKind.FormatString)
							{
								// 注意这里使用 IFormattable 来验证目标类型是否实现了格式化功能
								var obj = this._variables[token.Text] as IFormattable;

								value = obj == null ? this._variables[token.Text].ToString() : obj.ToString(nextToken.Text, null);
							}
							else value = this._variables[token.Text].ToString();
						}
						else value = this._variables[token.Text].ToString();

						result.Append(value);
						break;

					case TokenKind.Text:
						result.Append(token.Text);
						break;
				}
			}

			return result.ToString();
		}

		#endregion

		private void _EnterMode(LexerMode mode)
		{
			if (this._postion > 0)
			{
				var text = this._templateString.Substring(this._savedPostion, this._postion - this._savedPostion);
				var token = this._CreateToken(text);

				if (token != null)
				{
					this._tokens.Add(token);

					this._Save();
				}
			}

			this._lexerModes.Push(this._currentMode);
			this._currentMode = mode;
		}

		private void _LeaveMode()
		{
			if (this._postion > 0)
			{
				var text = this._templateString.Substring(this._savedPostion, this._postion - this._savedPostion);
				var token = this._CreateToken(text);

				if (token != null)
				{
					this._tokens.Add(token);

					this._Save();
				}
			}

			this._currentMode = this._lexerModes.Pop();
		}

		private Token _CreateToken(string text)
		{
			switch (this._currentMode)
			{
				case LexerMode.Label:
					return this._CreateToken(TokenKind.Label, text);

				case LexerMode.FormatString:
					return this._CreateToken(TokenKind.FormatString, text);

				default:
					return this._CreateToken(TokenKind.Text, text);
			}
		}

		private Token _CreateToken(TokenKind kind, string text)
		{
			if ((kind == TokenKind.Label || kind == TokenKind.FormatString) && String.IsNullOrEmpty(text)) return null;

			return new Token(kind, text ?? String.Empty, this._savedLine, this._savedColumn);
		}

		private void _Save()
		{
			this._savedLine = this._line;
			this._savedColumn = this._column;
			this._savedPostion = this._postion;
		}
	}
}
